import { controller } from "@/components/treeview/signals";
import { globalLookup } from "./global";
import { WaveContainerView } from "./wave-container-view";
import { isScope, isVariable } from "./utils";
import { BSON } from "bson";

export const recoverConfig = {
    /**
     * @type {Map<string, IRenderOption>} 其中的 string 是形如 `cpu.alu.add1` 这样的字符串
     */
    waves: new Map(),
    views: undefined
};

async function findRecoverFile(recoverJsonPath) {
    try {
        const response = await fetch(recoverJsonPath);
        if (!response.ok) {
            return undefined;
        }
        const recoverBuffer = await response.arrayBuffer();
        const buffer = Buffer.from(recoverBuffer);
        const recoverJson = BSON.deserialize(buffer);
        return recoverJson;
    } catch (error) {
        return undefined;
    }
}

/**
 * 
 * @param {string} inputFile 读取的 vcd 文件的路径
 */
export async function recoverFromInputFile(inputFile) {
    globalLookup.originFile = inputFile;
    console.log(inputFile);
    
    const recoverJsonPath = inputFile + '.view';
    const recoverJson = await findRecoverFile(recoverJsonPath);
    if (recoverJson) {
        const waves = recoverJson.waves;
        if (waves instanceof Array && waves.length > 0) {
            for (const wave of waves) {
                const name = wave.name;
                const option = wave.option;
                if (name === undefined || option === undefined || recoverConfig.waves.has(name)) {
                    continue;
                }
                recoverConfig.waves.set(name, option);
            }
        }
    }
}

class PrefixNode {
    constructor() {
        this.children = new Map();
        this.payload = undefined;
    }
}



/**
 * @description 恢复现场的函数，主要恢复两个变量 currentWiresRenderView 和 currentSignalRenderOptions
 * @param {TopModWireItem[]} topModules
 */
export function recoverSession(topModules) {
    const waves = recoverConfig.waves;
    if (waves.size === 0) {
        return;
    }

    // 利用前缀树 记忆化搜索
    for (const [name, option] of waves.entries()) {
        const result = names2wire(topModules, name);
        if (result) {
            // 找到了 进行渲染设置
            WaveContainerView.add(result);
            globalLookup.currentSignalRenderOptions.set(result.link, option);
            controller.lastSignal = result;
        } else {
            console.log('fail to recover ' + name);
        }
    }

    globalLookup.render();
}

/**
 * 
 * @param {TopModWireItem[]} topModules 
 * @param {string} name
 * @returns {WireItem | undefined} 
 */
function names2wire(topModules, name) {
    const deps = name.split('.');
    let layer = 0;
    let nodes = topModules;
    let result = undefined;
    let lastNode = undefined;

    while (layer < deps.length) {
        // 从当前 nodes 中匹配 className
        const targetName = deps[layer];
        let targetNode = nodes.filter(n => n.name == targetName)[0];
        if (targetNode === undefined) {
            break;
        }

        if (isScope(targetNode)) {
            nodes = targetNode.body;
        } else if (isVariable(targetNode)) {
            result = targetNode;
            if (result.parent === undefined) {
                result.parent = lastNode;
            }
        } else {
            break;
        }
        lastNode = targetNode;
        layer ++;
    }

    if (result !== undefined) {
        return result;
    }
    return undefined;
}

/**
 * 
 * @param {string} link 
 * @returns {string}
 */
function link2names(link) {
    const names = [];
    const signal = globalLookup.link2CurrentWires.get(link);
    if (signal) {
        let node = signal;
        while (node !== undefined) {
            names.push(node.name);
            node = node.parent;
        }
    }

    return names.reverse().join('.');
}

/**
 * @template T
 * @param {T} obj 
 * @returns {T}
 */
function stableClone(obj) {
    if (window.structuredClone) {
        return structuredClone(obj);
    }
    return JSON.parse(JSON.stringify(obj));
}

export function makeSaveViewPayload() {
    const waves = [];
    const views = [];

    for (const wire of globalLookup.currentWires) {
        const option = globalLookup.currentSignalRenderOptions.get(wire.link) || {};
        const name = link2names(wire.link);
        waves.push({ name, option });
    }
    for (const view of globalLookup.currentWiresRenderView) {
        // link 转换成 names
        const saveView = stableClone(view);
        const viewLink = view.signalInfo.link;
        saveView.signalInfo.link = link2names(viewLink);
        const parentLink = saveView.signalInfo.parentLink;
        if (parentLink !== undefined) {
            saveView.signalInfo.parentLink = parentLink.endsWith('-group') ? link2names(parentLink.slice(0, -6)) + '-group' : viewLink;
        }
        // children 只会有一层
        for (const child of saveView.children) {
            child.signalInfo.link = link2names(child.signalInfo.link);
        }
        views.push(saveView);
    }    
    return { waves, views };
}